/*---------------------------------------------------------------
 *	UARTEの動作確認:P0が送信側(P0TX_P1RX) (μT-Kernel 3.0用)
 *
 *	Copyright (C) 2022-2024 by T3 WG of TRON Forum
 *---------------------------------------------------------------*/
#include <tk/tkernel.h>
#include <tm/tmonitor.h>

//-------- UARTEのレジスタ定義 ----------------------------------
#define	UARTE(p, r)	(UARTE##p##_BASE + UARTE_##r)

#define	UARTE0_BASE	0x40002000	// コンソール用UART0と競合するので使用しない
#define	UARTE1_BASE	0x40028000	// こちらを使用する

#define	UARTE_TASKS_STARTRX		(0x000)
#define	UARTE_TASKS_STOPRX		(0x004)
#define	UARTE_TASKS_STARTTX		(0x008)
#define	UARTE_TASKS_STOPTX		(0x00C)
#define	UARTE_TASKS_FLUSHRX		(0x02C)

#define	UARTE_EVENTS_CTS		(0x100)
#define	UARTE_EVENTS_NCTS		(0x104)
#define	UARTE_EVENTS_RXDRDY		(0x108)
#define	UARTE_EVENTS_ENDRX		(0x110)
#define	UARTE_EVENTS_TXDRDY		(0x11C)
#define	UARTE_EVENTS_ENDTX		(0x120)
#define	UARTE_EVENTS_ERROR		(0x124)
#define	UARTE_EVENTS_RXTO		(0x144)

#define	UARTE_EVENTS_RXSTARTED	(0x14C)
#define	UARTE_EVENTS_TXSTARTED	(0x150)
#define	UARTE_EVENTS_TXSTOPPED	(0x158)

#define	UARTE_SHORTS			(0x200)
#define	UARTE_INTEN				(0x300)
#define	UARTE_INTENSET			(0x304)
#define	UARTE_INTENCLR			(0x308)
#define	UARTE_ERRORSRC			(0x480)
#define	UARTE_ENABLE			(0x500)
#define	UARTE_PSEL_RTS			(0x508)
#define	UARTE_PSEL_TXD			(0x50C)
#define	UARTE_PSEL_CTS			(0x510)
#define	UARTE_PSEL_RXD			(0x514)
#define	UARTE_BAUDRATE			(0x524)

#define	UARTE_RXD_PTR			(0x534)
#define	UARTE_RXD_MAXCNT		(0x538)
#define	UARTE_RXD_AMOUNT		(0x53C)
#define	UARTE_TXD_PTR			(0x544)
#define	UARTE_TXD_MAXCNT		(0x548)
#define	UARTE_TXD_AMOUNT		(0x54C)

#define	UARTE_CONFIG			(0x56C)

//-------- LED制御のためのGPIOの初期設定 ------------------------
LOCAL void led_init(void)
{
	// COL1..COL5とROW1..ROW5に接続されたGPIOピンを出力に設定
	out_w(GPIO(P0, PIN_CNF(28)), 1);	// GPIO P0.28(COL1に接続)
	out_w(GPIO(P0, PIN_CNF(11)), 1);	// GPIO P0.11(COL2に接続)
	out_w(GPIO(P0, PIN_CNF(31)), 1);	// GPIO P0.31(COL3に接続)
	out_w(GPIO(P1, PIN_CNF(05)), 1);	// GPIO P1.05(COL4に接続)
	out_w(GPIO(P0, PIN_CNF(30)), 1);	// GPIO P0.30(COL5に接続)
	out_w(GPIO(P0, PIN_CNF(21)), 1);	// GPIO P0.21(ROW1に接続)
	out_w(GPIO(P0, PIN_CNF(22)), 1);	// GPIO P0.22(ROW2に接続)
	out_w(GPIO(P0, PIN_CNF(15)), 1);	// GPIO P0.15(ROW3に接続)
	out_w(GPIO(P0, PIN_CNF(24)), 1);	// GPIO P0.24(ROW4に接続)
	out_w(GPIO(P0, PIN_CNF(19)), 1);	// GPIO P0.19(ROW5に接続)
}

//-------- GPIOの特定のピンpinの出力をval(0または1)に設定 -------
//	pinで指定された以外のピンの状態は変化しない
LOCAL void out_gpio_pin(UW port, UW pin, UW val)
{
	INT port_addr;

	if(port == 1){							// P1の場合
		if(val)
			port_addr = GPIO(P1, OUTSET);	// P1でvalが1の場合
		else
			port_addr = GPIO(P1, OUTCLR);	// P1でvalが0の場合

	} else {								// P0の場合
		if(val)
			port_addr = GPIO(P0, OUTSET);	// P0でvalが1の場合
		else
			port_addr = GPIO(P0, OUTCLR);	// P0でvalが0の場合
	}
	out_w(port_addr, (1 << pin));			// 指定のピンに1を出力
}

//-------- ボタンスイッチに対応するGPIO P0のピンを入力に設定 ----
LOCAL void btn_init(void)
{
	out_w(GPIO(P0, PIN_CNF(14)), 0);	// ボタンスイッチAはGPIO P0.14に接続
	out_w(GPIO(P0, PIN_CNF(23)), 0);	// ボタンスイッチBはGPIO P0.23に接続
}

//-------- ミリ秒単位の現在時刻の下位32ビットを返す -------------
LOCAL W cur_time(void)
{
	SYSTIM	tim;
	tk_get_tim(&tim);			// 現在時刻を取得
	return(tim.lo);
}

//---------------------------------------------------------------
// UARTEの送信(TX)と受信(RX)に使用するエッジコネクタ端子の選択(※L)
//	以下の2行のうち一方のみコメントを外す
#define		P0TX_P1RX	TRUE	// RING0(=P0)から送信、RING1(=P1)から受信
// #define	P1TX_P0RX	TRUE	// RING1(=P1)から送信、RING0(=P0)から受信

//-------- UARTE1の初期化とピン設定(※A) ------------------------
LOCAL void init_uarte1(INT pin_txd, INT pin_rxd)
{
	out_w(UARTE(1, EVENTS_ENDRX), 0);
	out_w(UARTE(1, EVENTS_ENDTX), 0);
	out_w(UARTE(1, EVENTS_ERROR), 0);
	out_w(UARTE(1, ERRORSRC), 0);

	out_w(UARTE(1, BAUDRATE), 0x00275000);		// 9600 baud(※B)
	// out_w(UARTE(1, BAUDRATE), 0x01D7E000);	// 115200 baud(※B)

	// Hardware flow control無し、パリティ無し、One stop bit
	out_w(UARTE(1, CONFIG), 0);

	out_w(UARTE(1, PSEL_TXD), pin_txd);	// TXDのGPIOピン設定(※C)
	out_w(UARTE(1, PSEL_RXD), pin_rxd);	// RXDのGPIOピン設定(※C)
	out_w(UARTE(1, ENABLE), 8);			// Enable UARTE(※D)
}

//-------- UARTE1の送信(※E) ------------------------------------
LOCAL void uarte1_tx(INT txcnt, UB *txadr)
{
	out_w(UARTE(1, TXD_PTR), (UW) txadr);	// 送信バッファ先頭アドレス
	out_w(UARTE(1, TXD_MAXCNT), txcnt);		// 送信するバイト数

	out_w(UARTE(1, EVENTS_ENDTX), 0);		// 送信終了フラグのクリア
	out_w(UARTE(1, TASKS_STARTTX), 1);		// 送信開始(※F)

	// 送信終了までポーリングで待ってから戻る(※G)
	while(! in_w(UARTE(1, EVENTS_ENDTX)))
		tk_dly_tsk(10);
}

//-------- UARTE1の受信(※H) ------------------------------------
LOCAL void uarte1_rx(INT rxcnt, UB *rxadr)
{
	out_w(UARTE(1, RXD_PTR), (UW) rxadr);	// 受信バッファ先頭アドレス
	out_w(UARTE(1, RXD_MAXCNT), rxcnt);		// 受信するバイト数

	out_w(UARTE(1, EVENTS_ENDRX), 0);		// 受信終了フラグのクリア
	out_w(UARTE(1, TASKS_STARTRX), 1);		// 受信開始(※J)

	// 受信終了までポーリングで待ってから戻る(※K)
	while(! in_w(UARTE(1, EVENTS_ENDRX)))
		tk_dly_tsk(10);
}

//---------------------------------------------------------------
static W	i_time;				// システム起動時刻

// UARTEの送信データと受信データを入れるバッファメモリ
#define	MAXBUF	100
static _UB	txbuf[MAXBUF];
static _UB	rxbuf[MAXBUF];

// 以前のボタンの状態を保持する変数
LOCAL BOOL prev_a = FALSE;
LOCAL BOOL prev_b = FALSE;

//---------------------------------------------------------------
// ボタンの状態変化のチェックと送信(※T)
LOCAL void check_btn(BOOL new_btn, BOOL *prev_btn, UB btnchr)
{
	UB actchr;
	if(new_btn ==  *prev_btn) return;	// 状態変化が無ければ何もせずに戻る

	*prev_btn = new_btn;	
	actchr = (new_btn ? 'p' : 'r');		// ボタンを押した時にp、離した時にr

	// 送信する文字列を生成してtxbufに書き込む(※V)
	tm_sprintf((UB*) txbuf, "%c%c_%06d\n", btnchr, actchr, cur_time() - i_time);
	
	// txbufに設定した10バイトのデータをUARTE1で送信(※U)
	uarte1_tx(10, (UB*) txbuf);
}

// ボタンスイッチの入力判定と送信側タスク(※R)
LOCAL void tx_task(INT stacd, void *exinf)
{
	for(;;){				// 約100msごとに永久に繰り返し
		UW gpio_p0in;
		BOOL btn_a, btn_b;

		// GPIO P0のINレジスタを読んでgpio_p0inに設定
		gpio_p0in = in_w(GPIO(P0, IN));

		btn_a = ((gpio_p0in & (1 << 14)) == 0); // P0.14が0の場合にTRUE(※S)
		btn_b = ((gpio_p0in & (1 << 23)) == 0); // P0.23が0の場合にTRUE(※S)

		check_btn(btn_a, &prev_a, 'A');	// ボタンAの変化のチェックと送信
		check_btn(btn_b, &prev_b, 'B');	// ボタンBの変化のチェックと送信

		tk_dly_tsk(100);				// 100msのディレイ
	}
}

// 受信したメッセージをLEDとコンソールに表示する受信側タスク(※W)
LOCAL void rx_task(INT stacd, void *exinf)
{
	for(;;){				// 永久に繰り返し

		// UARTE1からrxbufに10バイトのデータを受信(※X)
		uarte1_rx(10, (UB*) rxbuf);

		rxbuf[9] = '\0';					// 10バイト目に終端文字を設定
		tm_printf("%s\n", (UB*) rxbuf);		// コンソールへのメッセージ出力

		if(rxbuf[0] == 'A'){
			if(rxbuf[1] == 'p'){			// ボタンAがON("Ap")
				// ROW3のCOL1(D22)を点灯(ROW3=1,COL1=0)(※Y)
			    out_gpio_pin(0, 28, 0);     // GPIO P0.28-COL1に0を出力
			} else if(rxbuf[1] == 'r'){		// ボタンAがOFF("Ar")
				// ROW3のCOL1(D22)を消灯(ROW3=1,COL1=1)(※Z)
			    out_gpio_pin(0, 28, 1);     // GPIO P0.28-COL1に1を出力
			}
		} else if(rxbuf[0] == 'B'){
			if(rxbuf[1] == 'p'){			// ボタンBがON("Bp")
				// ROW3のCOL5(D30)を点灯(ROW3=1,COL5=0)
			    out_gpio_pin(0, 30, 0);     // GPIO P0.30-COL5に0を出力
			} else if(rxbuf[1] == 'r'){		// ボタンBがOFF("Br")
				// ROW3のCOL5(D30)を消灯(ROW3=1,COL5=1)
			    out_gpio_pin(0, 30, 1);     // GPIO P0.30-COL5に1を出力
			}
		}
	}
}

//---------------------------------------------------------------
// オブジェクトID番号
ID txtskid;				// 送信側タスク
ID rxtskid;				// 受信側タスク

// タスクの生成情報
const T_CTSK ctxtsk = {0, (TA_HLNG | TA_RNG3), &tx_task, 10, 1024, 0};
const T_CTSK crxtsk = {0, (TA_HLNG | TA_RNG3), &rx_task, 10, 1024, 0};

// UARTEの送信(TX)と受信(RX)に使用するエッジコネクタ端子の定義(※M)
#ifdef P0TX_P1RX
const INT tx_pin = 2;	// 送信をGPIO P0.02=端子P0(RING0)に設定
const INT rx_pin = 3;	// 受信をGPIO P0.03=端子P1(RING0)に設定
#endif

#ifdef P1TX_P0RX
const INT tx_pin = 3;	// 送信をGPIO P0.03=端子P1(RING0)に設定
const INT rx_pin = 2;	// 受信をGPIO P0.02=端子P0(RING0)に設定
#endif

//---------------------------------------------------------------
EXPORT void usermain( void ){
	i_time = cur_time();			// システム起動時刻を保存
	btn_init();						// ボタンスイッチ入力のためのGPIOの初期設定
	led_init();						// LED制御用のGPIO設定

	// GPIO P0.15-ROW3に1を出力してもROW3のLEDを点灯させないための設定(※N)
	out_gpio_pin(0, 28, 1);			// GPIO P0.28-COL1に1を出力
	out_gpio_pin(0, 11, 1);			// GPIO P0.11-COL2に1を出力
	out_gpio_pin(0, 31, 1);			// GPIO P0.31-COL3に1を出力
	out_gpio_pin(1,  5, 1);			// GPIO P1.05-COL4に1を出力
	out_gpio_pin(0, 30, 1);			// GPIO P0.30-COL5に1を出力

	// ROW3のLEDを点灯させるための設定(※N)
	out_gpio_pin(0, 15, 1);			// GPIO P0.15-ROW3に1を出力

	// 動作確認用にCOL3のLEDを点灯(※P)
	out_gpio_pin(0, 31, 0);			// GPIO P0.31-COL3に0を出力

	init_uarte1(tx_pin, rx_pin);	// UARTEの初期化と送受信用端子の指定

	txtskid = tk_cre_tsk(&ctxtsk);	// 送信側タスクの生成(※Q)
	rxtskid = tk_cre_tsk(&crxtsk);	// 受信側タスクの生成(※Q)
	tk_sta_tsk(txtskid, 0);			// 送信側タスクの起動(※Q)
	tk_sta_tsk(rxtskid, 0);			// 受信側タスクの起動(※Q)

	tk_slp_tsk(TMO_FEVR);			// 永久待ち
}
